# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The Query module

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 07:04:20
    `name`: query
    * @xxx [04-28-2023 07:04:34]: documentation for query
'''



from dataclasses import dataclass
import re
from select import select
from typing import Iterable, Union
from datetime import datetime
from datetime import timezone

import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
# import volent.settings as _settings
# from volent.Field import Field as _field
from volent.mixins import OrderedClass
from volent.query.WhereClause import WhereClause as _w
from volent.query.QueryParam import QueryParam as _param


@dataclass
class Query(metaclass=OrderedClass):
    main = None
    # database:_t.database_type = None
    model:_t.model_type = None
    '''A reference to the model instance this query is referencing.'''

    query_crud_type:str = None
    data:dict = None
    database:_t.database_type = None
    '''A reference to the database instance used to execute this query.'''

    _selects:Iterable[str] = None
    _wheres:Iterable[str] = None
    _count:bool = False
    _average:bool = False
    _sum:bool = False

    _where_clauses:Iterable[_t.where_clause_type] = None
    _params:Iterable[_t.query_param_type] = None



    _items_per_page:int=_settings.control.query.pagination.items_per_page
    '''How many items should be shown per page when the query is paginated.

    This directly corresponds to the "LIMIT" value.'''


    _page_number:int=_settings.control.query.pagination.default_page_number
    '''Which page of items should be returned.

    This is used in conjunction with the items_per_page to calculate the "OFFSET" value.
    '''

    def __init__(
        self,
        model:_t.model_type,
        data:dict=None,
        select_columns=None,
        update_columns=None
        ):
        '''
            Create a new query instance.

            ----------

            Arguments
            -------------------------
            `model` {model}
                The model instance that this query will act on.
            `data` {dict}
                A dictionary of data that is used in the query operation.
            `select_columns` {list}
                A list of column names or column instances that should be "selected" in a select query
            `update_columns` {list}
                A list of column names or column instances that should be updated in an update query


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:04:55
            `memberOf`: Query
            `version`: 1.0
            `method_name`: query
            * @xxx [04-28-2023 07:07:59]: documentation for query
        '''
        self.model = model
        self.data = data
        self.database = model.database
        self._params = []
        self._wheres = []
        self._where_clauses = []
        self._selects = []
        self._updates = {}

        self.__parse_col_data(select_columns,update_columns)
        # else:
        #     for col in update_columns:
        #         if isinstance(col,(list,tuple)):
        #             if len(col) == 2:
        #                 name,value = col
        #                 self.add_update(name,value)
            #     if len(col) == 1:
            #         col = col[0]
            # if isinstance(col,(str)):
            #     self.add_select(col)

    def __parse_col_data(
        self,
        select_columns:Iterable[Union[str,_t.column_type]],
        update_columns:Iterable[Union[str,_t.column_type]]
        ):
        '''
            Parse the select and update columns into this query for later use.

            Essentially take a bunch of column names/instances and store them in a useful way.
            ----------

            Arguments
            -------------------------
            `select_columns` {list}
                A list of column names or column instances that should be "selected" in a select query

            `update_columns` {list}
                A list of column names or column instances that should be updated in an update query

            Return
            ----------------------
            Doesn't return shit.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:08:04
            `memberOf`: Query
            `version`: 1.0
            `method_name`: __parse_col_data
            * @xxx [04-28-2023 07:10:20]: documentation for __parse_col_data
        '''
        if isinstance(select_columns,(str)):
            select_columns = c.arr.force_list(select_columns)
        if isinstance(select_columns,(list)):
            for col in select_columns:
                if isinstance(col,(list,tuple)):
                    if len(col) >= 2:
                        name,alias = col
                        self.add_select(name,alias)
                    if len(col) == 1:
                        col = col[0]
                if isinstance(col,(str)):
                    # print(f"adding column by name: {col}")
                    self.add_select(col)

        if isinstance(update_columns,(dict)):
            mt_col = self.model.get_column("modified_timestamp")
            if mt_col is not None:
                self.add_update("modified_timestamp",round(datetime.now(tz=timezone.utc).timestamp()))


            for k,v in update_columns.items():
                col = self.model.get_column(k)
                # @Mstep [IF] if the column is the primary key of the table
                if col.is_primary is True:
                    # @Mstep [] skip
                    continue
                # @Mstep [] add a new update clause to this query.
                self.add_update(k,v)

            for col in self.model.columns:
                # @Mstep [IF] if the column has an "on_update" value set.
                if col.on_update != _t.undefined:
                    # @Mstep [IF] if the column is not already being updated.
                    if col.name not in self._updates:
                        # @Mstep [] add a new update clause to this query.
                        self.add_update(col.name,col.on_update)

    @property
    def items_per_page(self):
        value = self._items_per_page
        if value is None:
            value = 100
            self._items_per_page = value
        return value

    @items_per_page.setter
    def items_per_page(self,value):
        self._items_per_page = value




    def filter_dict_by_columns(self,data:dict)->dict:
        '''
            Iterate a dictionary to remove key's that do not have a corresponding column in this query's model
            ----------

            Arguments
            -------------------------
            `data` {dict}
                The dictionary to filter.

            Return {dict}
            ----------------------
            The dictionary with only keys that match a column.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:10:41
            `memberOf`: Query
            `version`: 1.0
            `method_name`: filter_dict_by_columns
            * @xxx [04-28-2023 07:11:51]: documentation for filter_dict_by_columns
        '''
        output = {}
        for k,v in data.items():
            col = self.model.get_column(k)
            if col is not None:
                if v is None and col.nullable is False:
                    continue
                output[k] = v
        return output

    def filter_undefined_values(self,data:dict)->dict:
        '''Iterate a dictionary to remove any undefined values'''
        out = {}
        for k,v in data.items():
            if v == _t.undefined:
                continue
            out[k] = v
        return out

    def sort_dict_by_columns(self,data:dict)->dict:
        '''
            Sort the keys of dictionary to match the order of columns in this query's model.

            This is primarily useful for parameterized insert statements, where the order of the columns
            is critical to inserting the correct data.

            This will sort them into the same order that the columns were declared in.

            ----------

            Arguments
            -------------------------
            `data` {dict}
                The dictionary to sort.


            Return {dict}
            ----------------------
            The dictionary with all keys reordered to match the model's columns.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:11:58
            `memberOf`: Query
            `version`: 1.0
            `method_name`: sort_dict_by_columns
            * @xxx [04-28-2023 07:13:57]: documentation for sort_dict_by_columns
        '''
        output = {}
        for col in self.model.columns:
            if col.name in data:
                output[col.name] = data[col.name]
        return output



    @property
    def where_count(self):
        '''
            Get this Query's where_count

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-01-2024 16:56:43
            `@memberOf`: Query
            `@property`: where_count
        '''
        return len(self._where_clauses)

    # def add_where_obj(self,WhereClause:_t.where_clause_type):
    #     self._where_clauses.append(WhereClause)


    def add_param(self,key,value)->_t.query_param_type:
        p = _param(self,key,value)
        self._params.append(p)
        return p

    @property
    def serialized_params(self)->dict:
        '''
            Get a dictionary of all params associated to this query.

            The parameter names are serialized according to the flavor.

            Return {dict}
            ----------------------
            The serialized param dictionary

            Example
            ----------------------
            ```
            {
                "@track_number":34,
                "@title":"AlienBiscuits"
            }
            ```

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-02-2024 09:44:08
            `memberOf`: Query
            `version`: 1.0
            `method_name`: serialized_params
            * @xxx [04-02-2024 09:46:24]: documentation for serialized_params
        '''
        value = {}
        for p in self._params:
            value[p.name] = p.value
        return value

    def add_where(self,column_name,value,comparison):
        '''
            Add a where clause to this query.

            ----------

            Arguments
            -------------------------
            `column_name` {str}
                The name of the column that this clause applies to.
            `value` {any}
                The value that the column is compared to
            `comparison` {str}
                The comparison/operator to use.

            Return {type}
            ----------------------
            no returned value

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:14:41
            `memberOf`: Query
            `version`: 1.0
            `method_name`: add_where
            * @xxx [04-28-2023 07:19:24]: documentation for add_where
        '''
        # if value is None:
        #     value = "NULL"
        # # self.wheres.append(f"{column_name} {comparison} {value}")



        # if c.string.to_snake_case(comparison) in ["!","!=","isnt","isnot","is_not","<>"]:
        #     comparison = "is not"

        # data = {
        #     "column_name":column_name,
        #     "comparison":comparison,
        #     "value":value,
        #     "max_value":None,
        # }

        # if c.string.to_snake_case(comparison) in ["match","full_text_search"]:
        #     data['comparison'] = "match"
        #     data['column_name'] = c.arr.force_list(column_name)

        # if c.string.to_snake_case(comparison) in ["between"]:
        #     if isinstance(value,(list,tuple)):
        #         data['value'] = value[0]
        #         data['max_value'] = value[1]
        #     else:
        #         data['value'] = 0
        #         data['max_value'] = value

        # if c.string.to_snake_case(comparison) in ["in"]:
        #     value = c.arr.force_list(value)
        #     if isinstance(value,(list,tuple)):
        #         items = []
        #         for idx,x in enumerate(value):
        #             if isinstance(x,(str)):
        #                 # key = f"{column_name}_{idx}"
        #                 # items[key] = f"'{x}'"
        #                 items.append(x)


        #             if isinstance(x,(int,float)):
        #                 # key = f"{column_name}_{idx}"
        #                 items.append(f"{x}")
        #                 # items[key] = f"{x}"

        #         # str_list = ', '.join(items)

        #         data['value'] = items

        # # self._wheres.append(data)
        w = _w(self,column_name,value,comparison)
        self._where_clauses.append(w)
        return w



    @property
    def where_string(self)->str:
        '''
            Generate this query's where string.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 12-09-2022 15:45:29
            `@memberOf`: UpdateQuery
            `@property`: where_string
        '''
        value = self._where_clauses
        # self._params = {}
        if len(self._where_clauses) > 0:
            wheres = []
            for where in self._where_clauses:
                if where.is_or_child is True:
                    c.con.log(f"    Query.where_string : skipping or child: {where}","cyan")
                    continue

                single_where = where.where_string()

                wheres.append(single_where)
            wheres = ' AND '.join(wheres)
            value = f" WHERE {wheres}"
        else:
            value = ""
        return value

    def _format_query_params(self,sql:str,args:dict)->str:
        '''
            Format an SQL query's parameters to use the python named template format.

            This will only replace matches that have a corresponding key in the args dictionary.

            Parameters can begin with a dollar sign or colon.


            SELECT * from blackholes WHERE hash_id=$hash_id

            SELECT * from blackholes WHERE hash_id=%(hash_id)s

            ----------

            Arguments
            -------------------------
            `sql` {str}
                The sql string to format.

            `args` {dict}
                The dictionary of parameter values.


            Return {str}
            ----------------------
            The sql statement with parameters replaced.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-09-2022 10:39:24
            `memberOf`: MySQLDatabase
            `version`: 1.0
            `method_name`: _format_query_params
            * @xxx [12-09-2022 10:43:23]: documentation for _format_query_params
        '''
        return _format_query_params(sql,args)

    def _paginate_select_query(self,sql)->str:
        '''
            Apply a limit and offset value to a select query statement.


            Arguments
            -------------------------
            `sql` {str}
                The sql statement to modify

            Return {str}
            ----------------------
            The sql statement with a limit and offset value applied.

            If the limit/offset if invalid no pagination is added.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 12-14-2022 11:45:47
            `memberOf`: QueryBase
            `version`: 1.0
            `method_name`: __paginate_select_query
            * @xxx [12-14-2022 11:46:12]: documentation for __paginate_select_query
        '''
        return _paginate_select_query(sql,self.limit,self.offset)

    def copy_wheres(self,query:_t.query_type)->_t.query_type:
        '''
            Used to copy where clauses from one query to another

            Arguments
            -------------------------
            `query` {Query}
                The source query to copy where instances from.

            Return {Query}
            ----------------------
            returns this query instance.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-02-2024 09:39:58
            `memberOf`: Query
            `version`: 1.0
            `method_name`: copy_wheres
            * @xxx [04-02-2024 09:40:45]: documentation for copy_wheres
        '''
        wheres = query._where_clauses
        for w in wheres:
            column_name = w.column_name
            value = w.value
            max_value = w.max_value
            comparison = w.comparison

            if max_value is not None:
                value = [value,max_value]
            self.add_where(column_name=column_name,value=value,comparison=comparison)

        return self

    def add_where_from_wheres(self,wheres:Iterable[dict]):
        '''This is used to copy all where instances from one query to another.'''
        for w in wheres:
            column_name = w["column_name"]
            value = w["value"]
            max_value = w["max_value"]
            comparison = w["comparison"]

            if max_value is not None:
                value = [value,max_value]

            self.add_where(column_name,value,comparison)
        return self

    def add_where_from_where(self,**kwargs):
        column_name = c.obj.get_kwarg(["column_name"],None,(str),**kwargs)
        value = c.obj.get_kwarg(["value"],None,None,**kwargs)
        max_value = c.obj.get_kwarg(["max_value"],None,None,**kwargs)
        comparison = c.obj.get_kwarg(["comparison"],None,None,**kwargs)

        if max_value is not None:
            value = [value,max_value]

        self.add_where(column_name,value,comparison)

        return self

    def gather_wheres(self,where:_t.where_clause_type)->Iterable[_t.where_clause_type]:
        print(f"Query.gather_wheres ===================================")
        total_wheres = len(self._where_clauses)
        occ = where.or_child_count
        child_ids = range((total_wheres-1)-occ,total_wheres-1)
        for cid in child_ids:
            sw = self._where_clauses[cid]
            sw.or_parent_id = where.or_parent_id
            sw.is_or_child = True
            print(sw)
        print(f"Query.gather_wheres ===================================")

    def get_wheres_by_or_id(self,or_id:str)->Iterable[_t.where_clause_type]:
        out = []
        for w in self._where_clauses:
            if w.or_parent_id == or_id and w.is_or_child is True:
                out.append(w)
        return out

    def set_on_update_values(self,data:dict)->dict:
        '''
        Assign key/value to an array when there is a column with an on_update value and the
        key does not yet exist in the dict.
        '''
        out = data
        for col in self.model.columns:
            on_update = col.on_update
            if on_update == _t.undefined:
                continue

            # @Mstep [] Filter out onUpdates that match a where clause column
            skip = False
            for w in self._where_clauses:
                if w.column_name == col.name:
                    skip = True
                    break
            if skip is True:
                continue


            if col.name not in out:
                val = on_update
                if callable(val):
                    val = val()
                out[col.name] = val
        return out




def format_null(value):
    '''
        convert a None value to a "null" string.

        ----------

        Arguments
        -------------------------
        `value` {any}
            The value to test.

        Return {any}
        ----------------------
        "NULL" if the value is None, the original value otherwise.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 04-28-2023 07:20:50
        `memberOf`: Query
        `version`: 1.0
        `method_name`: format_null
        * @xxx [04-28-2023 07:22:36]: documentation for format_null
    '''
    if value is None:
        return "NULL"
    return value




def _paginate_select_query(sql:str,limit:int=None,offset:int=None)->str:
    '''
        Apply a limit and offset value to a select query statement.
        ----------

        Arguments
        -------------------------
        `sql` {str}
            The sql statement to modify

        [`limit`=None] {int}
            The limit to apply to the results.

        [`offset`=None] {int}
            The offset to apply to the results.


        Return {str}
        ----------------------
        The sql statement with a limit and offset value applied.

        If the limit/offset if invalid no pagination is added.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 12-09-2022 11:09:53
        `memberOf`: MySQLDatabase
        `version`: 1.0
        `method_name`: _paginate_select_query
        * @xxx [12-09-2022 11:11:52]: documentation for _paginate_select_query
    '''
    if limit is None and offset is None:
        return sql
    if isinstance(limit,(str)):
        limit = re.sub(r'[^0-9]',"",limit)
        if len(limit) == 0:
            return sql
        limit = int(limit)

    if isinstance(offset,(str)):
        offset = re.sub(r'[^0-9]',"",offset)
        if len(offset) == 0:
            offset = None
        else:
            offset = int(offset)

    if limit == 0:
        limit = 1

    if offset is not None:
        if offset < 1:
            offset = None

    sql = c.string.strip(sql,[";"],"right")
    sql = re.sub(r'limit\s*[0-9]*\s*(,|offset)\s*(:?[0-9\s]*)?',"",sql,re.MULTILINE | re.IGNORECASE)

    limit_string = f"LIMIT {limit}"
    offset_string = ""
    if offset is not None:
        offset_string = f"OFFSET {offset}"
    paginate = f"{limit_string} {offset_string}"
    sql = f"{sql} {paginate}"
    return sql




def _format_query_params(sql:str,args:dict)->str:
    '''
        Format an SQL query's parameters to use the python named template format.

        This will only replace matches that have a corresponding key in the args dictionary.

        Parameters can begin with a dollar sign or colon.


        SELECT * from blackholes WHERE hash_id=$hash_id

        SELECT * from blackholes WHERE hash_id=%(hash_id)s

        ----------

        Arguments
        -------------------------
        `sql` {str}
            The sql string to format.

        `args` {dict}
            The dictionary of parameter values.


        Return {str}
        ----------------------
        The sql statement with parameters replaced.

        Meta
        ----------
        `author`: Colemen Atwood
        `created`: 12-09-2022 10:39:24
        `memberOf`: MySQLDatabase
        `version`: 1.0
        `method_name`: _format_query_params
        * @xxx [12-09-2022 10:43:23]: documentation for _format_query_params
    '''
    if isinstance(args,(dict)) is False:
        return sql
    # args = sorted(args.items(), key=lambda x: x[1], reverse=True)
    # @Mstep [] get a list of the argument keys
    arg_keys = list(args.keys())
    # @Mstep [] sort the keys from largest to smallest.
    arg_keys.sort(key=len, reverse=True)
    # @Mstep [] iterate the argument keys.
    for k in arg_keys:
        if _settings.globe.flavor in ['sqlite']:
            sql = re.sub(fr'[$:]{k}',f"@{k}",sql)
        else:
    # for k,v in args.items():
            # @Mstep [] replace the key with the parameterized version.
            sql = re.sub(fr'[$:]{k}',f"%({k})s",sql)

    # matches = re.findall(r'[$:]([a-z_0-9]*)',sql,re.IGNORECASE)
    # if isinstance(matches,(list)):
    #     for match in matches:
    #         if match in args:
    #             sql = sql.replace(f"${match}",f"%({match})s")

    return sql




from typing import TYPE_CHECKING,TypeVar as _TypeVar


# ---------------------------------------------------------------------------- #
#                               TYPE DECLARATIONS                              #
# ---------------------------------------------------------------------------- #

_main_type = None
model_type = None
database_type = None
column_type = None
type_base_type = None
query_type = None
where_clause_type = None
query_param_type = None

relationship_type = None
schema_type = None
field_type = None
nested_field_type = None
unique_constraint_type = None
full_text_index_type = None

insert_query_type = None
select_query_type = None
update_query_type = None
delete_query_type = None


if TYPE_CHECKING:

    from volent.Volent import Volent as _m
    _main_type = _TypeVar('_main_type', bound=_m)

    from volent.Column import Column as _OXC0
    column_type = _TypeVar('column_type', bound=_OXC0)

    from volent.Model import Model as _HFem
    model_type = _TypeVar('model_type', bound=_HFem)

    from volent.Database import Database as _e8Ig
    database_type = _TypeVar('database_type', bound=_e8Ig)



    from volent.query.Insert import Insert as _wyFSbRdHe6gB
    insert_query_type = _TypeVar('insert_query_type', bound=_wyFSbRdHe6gB)

    from volent.query.Select import Select as _YTxzOsQXPA2r
    select_query_type = _TypeVar('select_query_type', bound=_YTxzOsQXPA2r)


    from volent.query.Update import Update as _L1hIZD4YCbjF
    update_query_type = _TypeVar('update_query_type', bound=_L1hIZD4YCbjF)

    from volent.query.Delete import Delete as _FoUnRQwSOBhN
    delete_query_type = _TypeVar('delete_query_type', bound=_FoUnRQwSOBhN)



    from volent.data_types.TypeBase import TypeBase as _NZwn
    type_base_type = _TypeVar('type_base_type', bound=_NZwn)

    from volent.Relationship import Relationship as _gt25
    relationship_type = _TypeVar('relationship_type', bound=_gt25)

    from volent.query.WhereClause import WhereClause as _wqWn4l7D3AOh
    where_clause_type = _TypeVar('where_clause_type', bound=_wqWn4l7D3AOh)


    from volent.query.QueryParam import QueryParam as _MXvEFm2oA9Vp
    query_param_type = _TypeVar('query_param_type', bound=_MXvEFm2oA9Vp)


    from volent.Schema import Schema as _gmhq
    schema_type = _TypeVar('schema_type', bound=_gmhq)

    from volent.Field import Field as _zoTC
    field_type = _TypeVar('field_type', bound=_zoTC)

    from volent.NestedField import NestedField as _mUDo
    nested_field_type = _TypeVar('nested_field_type', bound=_mUDo)

    from volent.UniqueConstraint import UniqueConstraint as _xbst
    unique_constraint_type = _TypeVar('unique_constraint_type', bound=_xbst)


    from volent.query.Query import Query as _PsBw
    query_type = _TypeVar('query_type', bound=_PsBw)

    from volent.FullTextIndex import FullTextIndex as _eMkA
    full_text_index_type = _TypeVar('full_text_index_type', bound=_eMkA)


class NoDefault:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value

class Undefined:
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return self.value

    # def __repr__(self) -> str:
    #     return f"<{self.__class__.__name__} : {self.value}>"
no_default = NoDefault("__NO_DEFAULT_PROVIDED__")
undefined = Undefined("__UNDEFINED_VALUE__")









import os
from glob import glob
from setuptools import setup, find_packages
import colemen_utilities.build_utils.general as _gen


VERSION='0.1.8'
DESCRIPTION = 'volent'
LONG_DESCRIPTION = 'None'


_root_path = f"{os.getcwd()}/volent"
PY_MODULES = _gen.list_py_modules(
    _root_path,
    additions=['main']
)
_gen.purge_dist()

DATA_FILES = [
    # ('Lib/site-packages/volent/documentation', glob('documentation/*.md')),
    # ('where the files will be installed', glob('where to get the files to include')),
]

# Setting up
setup(
    name="colemen_volent",
    version=VERSION,
    author="Colemen Mitchell Atwood",
    author_email="<atwoodcolemen@gmail.com>",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    py_modules=PY_MODULES,
    # add any additional packages that
    # need to be installed along with your package. Eg: 'caer'
    install_requires=[
        'colemen_utils',
    ],

    keywords=['python'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Education",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 3",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
from typing import Iterable
import colemen_utils as c
import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.data_types.TypeBase import TypeBase as _type_base

@dataclass
class EncodedPrimary(_type_base):

    def __init__(self,data_length:int=None) -> None:
        super().__init__(data_length)
        self.min_data_length = 0
        self.max_data_length = 65535
        self.sql_type_name = "BIGINT"
        self.python_data_type = (int,str)
        self.open_api_data_type = "string"

        self._validate_data_length(data_length)

    def serializer(self,value,field_name:str=None):
        print(f"field_name:{field_name}")
        if isinstance(value,(str)):
            return self.deserialized_value(value,field_name)
        if isinstance(value,int):
            return self.serialized_value(value,field_name)

    def serialized_value(self,value,field_name:str=None):
        if isinstance(value,(str)):
            return value
        return c.string.string_encode_int(value)

    def deserialized_value(self,value,field_name:str=None):
        if isinstance(value,(int)):
            # return value
            raise ValidationError(f"Invalid ID provided {value}",field_name)

        val = c.string.string_decode_int(value)
        if isinstance(val,(int)):
            return val
        raise ValidationError(f"Invalid ID provided {value}",field_name=field_name)


    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} : {self.data_length}>"


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

# from dataclasses import dataclass
from typing import Iterable, OrderedDict, Sized, Union
# from collections import OrderedDict
# from abc import ABC, abstractmethod

from datetime import datetime
from datetime import timezone



import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator

class FutureUnixDate(Validator):

    field_name = None
    message_min = "Date is in the past."
    message_max = "Date is too far in the future."


    def __init__(
        self,
        max_days=None,
        ):

        self.max_days = max_days

    def _format_error(self, value: Sized, message: str) -> str:
        if self.field_name is not None:
            message = message[0].lower() + message[1:]
            message = f"{self.field_name} {message}"
        return message.format(
            input=value, min=self.min, max=self.max, equal=self.equal
        )


    def _repr_args(self) -> str:
        return f"max_days={self.max_days!r}"


    def __call__(self,value:int,field_name:str=None)->Sized:
        cur = round(datetime.now(tz=timezone.utc).timestamp())
        if value < cur:
            raise ValidationError(self.message_min,field_name)
        if isinstance(self.max_days,(int)):
            max = cur + (self.max_days * 86400)
            if value > max:
                raise ValidationError(self.message_max,field_name)

        return value


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import


from dataclasses import dataclass
from time import time
from typing import Iterable,OrderedDict, Union


import colemen_utils as c


import volent.settings.types as _t
import volent.settings as _settings
from volent.Field import Field as _field
# from volent.Relationship import Relationship as _relationship
# from volent.UniqueConstraint import UniqueConstraint as _uniqueConstraint
from volent.query.Query import Query


@dataclass
class Insert(Query):
    def __init__(self,model:_t.model_type,data:Union[dict,list]) -> None:
        self._inserts = {}
        self._multi_inserts = []
        self._is_multi = False
        self._unique_columns = []
        super().__init__(model,data)

    @property
    def query(self):
        if isinstance(self.data,(dict,list)) is False:
            raise ValueError("Data should be a dictionary or list.")
        data = self._inserts_from_data()
        if self._is_multi is True:
            return self._multi_query
        # data = self.data
        if len(list(self._inserts.keys())) == 0:
            return (False,False)

        template = f"""INSERT INTO {self.model.quoted_name} ({self.column_list_string}) VALUES ({self.placeholder_string})"""
        # data_tuple = data.values()
        # print(f"template: {template}")
        # print(f"data_tuple: {data_tuple}")
        return (template,data)

    @property
    def placeholder_string(self)->str:
        '''
            Get this Insert's placeholder_string

            This is the list of value placeholders used to parameterize the query.

            "%(beep)s,%(boop)s"

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-26-2023 11:15:02
            `@memberOf`: Insert
            `@property`: placeholder_string
        '''
        if self._is_multi is True:
            master_ph = []
            ph_data = {}
            for idx,row in enumerate(self._multi_inserts):
                ph = []
                for k,v in row.items():
                    key = f"{k}{idx}"
                    ph_key = f"%({key})s"
                    if _settings.globe.flavor in ['sqlite']:
                        ph_key = f"@{key}"
                    ph.append(ph_key)
                    ph_data[key] = v
                ph_string = ', '.join(ph)
                master_ph.append(f"({ph_string})")
            master_ph = ', '.join(master_ph)
            return (master_ph,ph_data)

        ph = []
        for k in list(self._inserts.keys()):
            if _settings.globe.flavor in ['sqlite']:
                ph.append(f"@{k}")
            else:
                ph.append(f"%({k})s")
        return ', '.join(ph)

    @property
    def column_list_string(self):
        '''
            Get this Insert's column_list_string

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-26-2023 11:13:45
            `@memberOf`: Insert
            `@property`: column_list_string
        '''
        if _settings.globe.flavor in ['sqlite']:
            if self._is_multi is True:
                return ','.join([f'"{x}"' for x in self._unique_columns])
            cols = list(self._inserts.keys())
            return ', '.join([f'"{x}"' for x in cols])
        if self._is_multi is True:
            return ','.join(self._unique_columns)
        return ', '.join(list(self._inserts.keys()))

    # @property
    # def query(self):
    #     data = self.data

    #     if isinstance(data,(dict)) is False:
    #         raise ValueError("Data should be a dictionary.")

    #     data = self.filter_dict_by_columns(data)
    #     if len(list(data.keys())) == 0:
    #         return (False,False)
    #     column_list = ', '.join(list(data.keys()))
    #     # total_keys = len(list(data.keys()))
    #     ph = []
    #     for k in list(data.keys()):
    #         ph.append(f"%({k})s")
    #     placeholders = ', '.join(ph)
    #     template = f"""INSERT INTO {self.model.quoted_name} ({column_list}) VALUES ({placeholders})"""
    #     # data_tuple = data.values()
    #     # print(f"template: {template}")
    #     # print(f"data_tuple: {data_tuple}")
    #     return (template,data)
    #     # return (template,data_tuple)

    @property
    def _multi_query(self):
        '''
            Get this Insert's _multi_query

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-27-2023 09:31:16
            `@memberOf`: Insert
            `@property`: _multi_query
        '''

        place_holders,data = self.placeholder_string

        template = f"""INSERT INTO {self.model.quoted_name} ({self.column_list_string}) VALUES {place_holders}"""
        # print(f"==================================================================")
        # print(data)
        # print(f"==================================================================\n\n")
        
        return (template,data)

    def _inserts_from_data(self):
        # print(f"insert._inserts_from_data")
        data = self.data
        # @Mstep [IF] if the data is a list.
        if isinstance(data,(list)):
            self._is_multi = True
            self._multi_inserts = []


            # @Mstep [LOOP] iterate the list of datas
            for d in data:
                # @Mstep [IF] if the element is not a dictionary
                if isinstance(d,(dict)) is False:
                    # @Mstep [] skip that shit.
                    continue
                d_insert = {}

                # @Mstep [] filter and sort the element by the columns of the model.
                d = self.sort_dict_by_columns(self.filter_dict_by_columns(d))
                # @Mstep [IF] if the filtered dict has no keys
                if len(d.keys()) == 0:
                    # @Mstep [] skip this element.
                    continue


                # @Mstep [loop] Iterate the items in the dict.
                for k,v in d.items():

                    # @Mstep [IF] if the value is not undefined.
                    if v != _t.undefined:
                        self.__add_unique_coluumn(k)
                        # @Mstep [] add the key and value to the d_insert dict.
                        d_insert[k] = v
                # @Mstep [LOOP] iterate the model columns.
                for col in self.model.columns:
                    # @Mstep [IF] if thhe column has an on_insert value.
                    if col.on_insert != _t.undefined:
                        # print(f"On insert column found : {col.name}")
                        # @Mstep [IF] if the column is not already being set.
                        if col.name not in d_insert:
                            # @Mstep [] set the key and value on d_insert.
                            d_insert[col.name] = col.on_insert
                            self.__add_unique_coluumn(col.name)
                # @Mstep [] append the d_insert dictionary to self._multi_inserts.
                self._multi_inserts.append(d_insert)

            # @Mstep [RETURN] return the _multi_inserts list.
            return self._multi_inserts





        data = self.filter_dict_by_columns(data)
        # print(f"data: {data}")
        for k,v in data.items():
            if v != _t.undefined:
                self.add_insert(k,v)

        for col in self.model.columns:
            # print(f"col.on_insert: {col.on_insert}")
            if col.on_insert != _t.undefined:
            # if isinstance(col.on_insert,(_t.undefined)) is False:
                # print(f"On insert column found : {col.name}")
                if col.name not in self._inserts:
                    self.add_insert(col.name,col.on_insert)
        return self._inserts

    def add_insert(self,column_name,value):
        self._inserts[column_name] = value

    def __add_unique_coluumn(self,name):
        '''Add a name to the unique columns list if it is not already apart.'''
        if name not in self._unique_columns:
            self._unique_columns.append(name)


    def execute(self,return_result:bool=False,foreign_key_checks:bool=True)->Union[bool,dict,int]:
        '''
            Execute the insert operation on the model.

            ----------

            Arguments
            -------------------------
            [`return_result`=False] {bool}
                If True, it will run a select query to gather the newly inserted row.
                If False, it will return only the ID of the newly inserted row.

            [`foreign_key_checks`=True] {bool}
                If False, the query will not perform foreign key checks before executing the insertion.
                Be careful, this can cause integrity issues and will NOT warn you about it.


            Return {bool,dict,int}
            ----------------------
            The id of the inserted row or the entire row if return_result is True.



            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-28-2023 07:34:05
            `memberOf`: Delete
            `version`: 1.0
            `method_name`: execute
            * @xxx [04-28-2023 07:40:57]: documentation for execute
        '''
        sql,args= self.query
        if sql is False:
            return False

        # print(f"volent.query.insert.sql:{sql}")
        # print(f"volent.query.insert.args:{args}")
        start = time()
        # @Mstep [] execute the insert query.
        result = self.database.run(sql,args,foreign_key_checks=foreign_key_checks)
        # @Mstep [IF] if the query was successful.
        if result is True and return_result is True:
            # @Mstep [] get the id of the inserted role.
            result = self.database.last_id()
            self.model.primary_column.value = result
            self.model._saved = True
            result = self.model.select().is_(self.model.primary_column.name,result).execute()
        if result is True and return_result is False:
            result = self.database.last_id()


        # print(f"sql: {sql}")
        # print(f"args: {args}")
        # print(f"insert execute duration: {time() - start} : {result}")
        return result



# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import
'''
    The main module of the volent library.

    ----------

    Meta
    ----------
    `author`: Colemen Atwood
    `created`: 04-28-2023 07:03:18
    `name`: volent
    * @xxx [04-28-2023 07:03:34]: documentation for volent
'''



import os
from typing import Iterable
import colemen_utils as c
import volent.settings as _settings
from volent.Model import Model as Model
from volent.Schema import Schema as Schema
from volent.Field import Field as Field
from volent.NestedField import NestedField as NestedField
from volent.Column import Column as Column
from volent.Relationship import Relationship as Relationship
from volent.UniqueConstraint import UniqueConstraint as UniqueConstraint
from volent.FullTextIndex import FullTextIndex as FullTextIndex
from volent.Database import Database as _Database
import volent.validate as validators
import volent.exceptions as exc
import volent.data_types as data_types
import volent.settings.types as _t
from volent.mixins.DatabaseConnection import DatabaseConnection as _DatabaseConnection
from volent.mixins.MySQLGeneratorMixin import MySQLGeneratorMixin as _MySQLGeneratorMixin
from volent.settings.types import no_default as noDefault
from volent.settings.types import undefined

class Volent(_DatabaseConnection,_MySQLGeneratorMixin):
    def __init__(
        self,
        index_existing_tables:bool=False
        ):
        self.models:Iterable[_t.model_type] = []
        '''A list of all registered model instances'''

        self.model_names:Iterable[str] = []
        '''A list of all registered model/table names'''

        self.unique_constraints:Iterable[_t.unique_constraint_type] = []
        '''A list of all registered unique constraint instances'''

        self.uq_names:Iterable[str] = []
        '''A list of all registered unique constraint names'''

        self.relationships:Iterable[_t.relationship_type] = []
        '''A list of all registered relationship instances'''

        self.relationship_names:Iterable[str] = []
        '''A list of all registered relationship names'''

        self.columns:Iterable[_t.column_type] = []
        '''A list of all registered column instances'''

        self._databases = []
        self.force_single_db = False

        self._index_existing_table:bool = index_existing_tables
        _settings.globe.Volent = self
        super().__init__()
        # self.set_defaults()

    # def set_defaults(self):
    #     self.settings = c.file.import_project_settings("volent.settings.json")

    def master(self):
        print("master")

    def summary(self,file_path:str=None)->dict:
        '''
            Generate a dictionary of summary data for all databases.

            ----------

            Arguments
            -------------------------
            [`file_path`=None] {str}
                The path to save the summary to, if not provided, it is saved to the working directory.

            Return {dict}
            ----------------------
            returns the summary dictionary

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 08:05:45
            `memberOf`: Volent
            `version`: 1.0
            `method_name`: summary
            * @xxx [04-25-2023 08:26:06]: documentation for summary
        '''
        if file_path is None:
            file_path = f"{os.getcwd()}/summary.json"

        data = {
            "databases":[x.summary for x in self.databases]
        }
        c.file.writer.to_json(file_path,data)
        return data


    def start(self):
        '''
            Initiate the models, determine their relationships and connect to the databse.

            Raises a runtime error if it fails to connect to the database.

            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 08:04:33
            `memberOf`: Volent
            `version`: 1.0
            `method_name`: start
            * @TODO []: documentation for start
        '''
        # self.models = []
        # self.model_names = []
        self.unique_constraints = []
        self.uq_names = []
        self.relationships = []
        self.relationship_names = []
        self.columns = []
        self._databases = []
        
        self._deep_registration()
        self._determine_relationships()
        self._generate_databases()
        # result = self.connect()
        if self.connect() is False:
            raise RuntimeError("Failed to connect to the database.")
        else:
            if self._index_existing_table is True:
                for db in self.databases:
                    db.existing_tables()


    def register(self,instance):
        '''
            Used internally for registration and mapping of the database structure.

            ----------

            Arguments
            -------------------------
            `instance` {any}
                The entity to register.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 08:26:17
            `memberOf`: Volent
            `version`: 1.0
            `method_name`: register
            * @xxx [04-25-2023 08:26:56]: documentation for register
        '''
        if isinstance(instance,Model):
            # c.con.log(f"Registering New Model: {instance.name}","info")
            if instance.model_name in self.model_names:
                raise ValueError(f"Duplicate Name for Model : {instance.model_name}")
            if isinstance(self.force_single_db,(str)):
                instance._database_name = self.force_single_db
            self.models.append(instance)
            self.model_names.append(instance.model_name)


        if isinstance(instance,UniqueConstraint):
            # c.con.log(f"Register Unique Constraint: {instance.name}","info")
            if instance.name in self.uq_names:
                raise ValueError(f"Duplicate Name for Unique Constraint : {instance.name}")
            self.unique_constraints.append(instance)
            self.uq_names.append(instance.name)

        if isinstance(instance,Relationship):
            # c.con.log(f"Register Relationship: {instance.name}","info")
            self.relationships.append(instance)
            self.relationship_names.append(instance.name)

        if isinstance(instance,Column):
            # c.con.log(f"Register Column: {instance.name}","info")
            self.columns.append(instance)

    def get_column(self,name)->_t.column_type:
        '''
            Retrieve a column by its name.

            "database.table.column"

            ----------

            Arguments
            -------------------------
            `name` {str}
                The dot notation name.

                {database}.{table}.{column}

                {table}.{column}

            Return {column,None}
            ----------------------
            The column if it can be found, None otherwise

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-24-2023 08:30:03
            `memberOf`: BlackViper
            `version`: 1.0
            `method_name`: get_column
            * @xxx [03-24-2023 08:31:40]: documentation for get_column
        '''
        name = c.string.strip_excessive_chars(name,["."])
        nl = name.split(".")
        if len(nl) == 3:
            db_name = nl[0]
            mdl_name = nl[1]
            col_name = nl[2]
            db = self.get_database(db_name)
            if db is None:
                raise ValueError(f"Failed to locate column from {name}, Could not find database {db_name}")
            model = db.get_model(mdl_name)
            if model is None:
                raise ValueError(f"Failed to locate column from {name} - Could not find model {mdl_name}")
            column = model.get_column(col_name)
            if column is None:
                raise ValueError(f"Failed to locate column from {name} - Could not find column {col_name}")
            return column
        if len(nl) == 2:
            mdl_name = nl[0]
            col_name = nl[1]
            model = self.get_model(mdl_name)
            if model is None:
                raise ValueError(f"Failed to locate column from {name} - Could not find model {mdl_name}")
            column = model.get_column(col_name)
            if column is None:
                raise ValueError(f"Failed to locate column from {name} - Could not find column {col_name}")
            return column

    def get_model(self,name:str)->_t.model_type:
        '''
            Retrieve a model by searching for its name.

            ----------

            Arguments
            -------------------------
            `name` {str}
                The name of the model to search for.

                This can also be a dot path: database_name.table_name


            Return {model}
            ----------------------
            The model instance if successfull, None otherwise.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 08:27:11
            `memberOf`: Volent
            `version`: 1.0
            `method_name`: get_model
            * @xxx [04-25-2023 08:28:14]: documentation for get_model
        '''
        name = c.string.strip_excessive_chars(name,["."])
        nl = name.split(".")
        if len(nl) == 2:
            name = nl[0]
        for md in self.models:
            if md.model_name == name:
                # print(f"---------------------------------name: {name}")
                return md

        return None

    # def get_base_model(self,)

    def _deep_registration(self):
        '''This will iterate all models to have their sub entities register with this instance.'''
        c.con.log("Initiating Deep Registration","info")
        for mdl in self.models:
            mdl.volent_register_subs()

    def _determine_relationships(self):
        c.con.log("Determining relationships between tables","info")

        # @Mstep [LOOP] iterate all relationship instances.
        # for rel in self.relationships:
        #     # @Mstep [] Attempt to retrieve the parent
        #     parent_string = rel.parent
        #     pm = self.get_model(parent_string)
        #     if pm is not None:
        #         print(f"    parent_string:{parent_string}")
        #         rel.parent_model = pm
        #     else:
        #         c.con.log(f"Failed to locate parent model for relationship {rel.name} from dot_path: {parent_string}","red")

        #     pc = self.get_column(parent_string)
        #     if pc is not None:
        #         rel.parent_column = pc



    @property
    def force_single_db(self):
        '''
            If this is a string, the __database_name__ of all models will be overridden to use that database.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 05-08-2023 13:29:36
            `@memberOf`: Volent
            `@property`: force_single_db
        '''
        value = self._force_single_db
        return value
    
    @force_single_db.setter
    def force_single_db(self,value):
        '''
            Set the Volent's force_single_db property

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 05-08-2023 13:29:50
            `@memberOf`: Volent
            `@property`: force_single_db
        '''
        self._force_single_db = value
        # for m in self.models:
        #     m._database_name = value



    # ---------------------------------------------------------------------------- #
    #                                   DATABASES                                  #
    # ---------------------------------------------------------------------------- #

    @property
    def databases(self)->Iterable[_t.database_type]:
        '''
            Get a list of databases being managed by this instance.

            The result is None if there are no databases.

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-18-2023 08:16:56
            `@memberOf`: Volent
            `@property`: databases
        '''
        value = self._databases
        if len(value) == 0:
            value = None
        return value

    def _generate_databases(self):
        '''
            Locate all databases by iterating the models and create an instance for each unique name.


            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-18-2023 08:13:40
            `memberOf`: Volent
            `version`: 1.0
            `method_name`: _generate_databases
            * @xxx [04-18-2023 08:16:10]: documentation for _generate_databases
        '''
        self._databases = []
        names = self.get_databases_with_creds()
        if len(names) == 0:
            c.con.log("No Databases were specified in the models.","warning")
            return None

        for db in names:
            # @Mstep [] instantiate the database.
            dbi = _Database(self,db)
            # @Mstep [] have the database locate its associated tables.
            dbi.register_models()
            # @Mstep [] add the database instance to the databases list.
            self._databases.append(dbi)

    @property
    def database_names(self)->Iterable[str]:
        '''
            Get this Volent's database_names

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-17-2023 15:47:21
            `@memberOf`: Volent
            `@property`: database_names
        '''
        db_names = []
        for m in self.models:
            db_name = m.database_name
            if isinstance(db_name,(str)):
                if db_name not in db_names:
                    db_names.append(db_name)
        return db_names













    def generate_sql(self,file_path:str=None,drops:bool=True)->str:
        '''
            Generate the master SQL file for the database.

            ----------

            Arguments
            -------------------------
            [`file_path`=None] {str}
                The path to save the SQL to, it will default to the cwd/master.sql

            [`drops`=True] {bool}
                If False, it will not drop existing tables before creating them.

            Return {str}
            ----------------------
            The master SQL content as a string.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-25-2023 08:29:02
            `memberOf`: Volent
            `version`: 1.0
            `method_name`: generate_sql
            * @TODO []: documentation for generate_sql
        '''
        if file_path is None:
            file_path = f"{os.getcwd()}/master.sql"
        sql = self.master_sql(file_path,drops)
        return sql

    def clean_slate(self):
        '''If this is an sqlite database, this will delete the ENTIRE database.'''
        for db in self.databases:
            db.connection.close()
            file_name = f"{db.name}.db"
            print(f"file_name:{file_name}")
            c.file.delete(file_name)

        self.start()
        self.create_all()


    def drop_all_tables(self):
        '''
            Be Fucking careful.
            ----------

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 04-18-2023 09:40:21
            `memberOf`: Volent
            `version`: 1.0
            `method_name`: drop_all
            * @TODO []: documentation for drop_all
        '''
        # print(f"DROPPING ALL TABLES")
        for db in self.databases:
            # print(f"    db:{db.name}")
            db.drop_tables()
            db.drop_model_tables()

    def create_all(self,force=False):
        '''
            Create all schemas and tables in the database.
            
            ----------

            Arguments
            -------------------------
            [`force`=bool] {False}
                If True, this will drop the tables then recreate them

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-25-2023 10:57:44
            `memberOf`: BlackViper
            `version`: 1.0
            `method_name`: create_all
            * @TODO []: documentation for create_all
        '''

        # @Mstep [loop] iterate all registered databases
        if _settings.globe.flavor == "mysql":
            for db in self.databases:
                # @Mstep [IF] if force is True
                if force is True:
                    # @Mstep [] drop the database schema
                    db.run(db.drop_statement)
                # @Mstep [] create the database schema
                db.run(db.create_statement,foreign_key_checks=False)

                for mdl in db.models:
                    if force is True:
                        db.run(mdl.drop_statement)
                    db.run(mdl.create_statement,foreign_key_checks=False)
        if _settings.globe.flavor == "sqlite":
            for db in self.databases:

                for mdl in db.models:
                    # print(f"creating table: {mdl.model_name}")
                    if force is True:
                        db.run(mdl.drop_statement)
                    # print(mdl.create_statement)
                    db.run(mdl.create_statement,foreign_key_checks=False)
                    # print(f"\n\n")

if __name__ == '__main__':
    m = Volent()
    m.master()




"""API tests"""
# pylint: disable=invalid-name

from dataclasses import dataclass
from datetime import datetime
from datetime import timezone


import colemen_utils as c
import volent

# import volent.Volent as volent
from volent.Volent import Schema,Field,validators as _v
from volent.data_types import String,Integer,EncodedPrimary
import volent.validate as valid



class TestValidators:
    """TEST Schema Validation of dictionaries"""

    def test_email(self):
        """Confirm that we can validate an email address."""
        v = valid.Email()
        email = c.rand.email()
        r = v(email,"email")
        assert r == email

        email = c.rand.abstract_name()
        try:
            v(email,"email")
        except Exception as e:
            return
        assert False

    def test_equal(self):
        """Confirm that the equal validator functions properly."""
        value = c.rand.rand()
        v = valid.Equal(value)
        r = v(value,"value")
        assert r == value

        incorrect_value = c.rand.rand()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return
        assert False

    def test_future_unix_timestamp(self):
        """Confirm that we can validate a future unix timestamp."""
        ct = round(datetime.now(tz=timezone.utc).timestamp())
        value = ct + (86400 * c.rand.number(1,10))
        v = valid.FutureUnixDate(value)
        r = v(value,"value")
        assert r == value

        incorrect_value = ct - (86400 * c.rand.number(1,10))
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return
        assert False

    def test_ip_address_ipv4(self):
        """Confirm that we can validate an ipv4 address."""
        value = c.rand.ip_address()
        v = valid.IpAddress()
        r = v(value,"value")
        assert r == value

        incorrect_value = c.rand.rand()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return
        assert False

    def test_ip_address_ipv6(self):
        """Confirm that we can validate an ipv6 address."""
        value = c.rand.ip_address_ipv6()
        v = valid.IpAddress()
        r = v(value,"value")
        assert r == value

        incorrect_value = c.rand.rand()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return
        assert False

    def test_length_range(self):
        """Confirm that we can validate that a string's length is within a range."""
        min = 12
        max = 14
        value = c.rand.rand(min)
        v = valid.Length(min,max)
        r = v(value,"value")
        assert r == value



        # -------------------------- Test below the minimum -------------------------- #
        incorrect_value = c.rand.rand(min - c.rand.number(1,10))
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return


        # -------------------------- Test above the maximum -------------------------- #
        incorrect_value = c.rand.rand(max + c.rand.number(1,10))
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return



        assert False

    def test_length_min(self):
        """Confirm that we can validate a strings length is greater than a minimum"""
        min = 12
        # max = 14
        value = c.rand.rand(min)
        v = valid.Length(min)
        r = v(value,"value")
        assert r == value



        # -------------------------- Test less than the minimum -------------------------- #
        incorrect_value = c.rand.rand(min - c.rand.number(1,10))
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return



        assert False

    def test_length_max(self):
        """Confirm that we can validate a strings length is less than a maximum"""
        # min = 12
        max = 14
        value = c.rand.rand(max)
        v = valid.Length(max=max)
        r = v(value,"value")
        assert r == value



        # -------------------------- Test greater than the maximum -------------------------- #
        incorrect_value = c.rand.rand(max + c.rand.number(1,10))
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return



        assert False

    def test_length_equal(self):
        """Confirm that we can validate a strings length is equal to a value"""
        # min = 12
        max = 8
        value = c.rand.rand(max)
        v = valid.Length(equal=max)
        r = v(value,"value")
        assert r == value



        # -------------------------- Test greater than the maximum -------------------------- #
        incorrect_value = c.rand.rand(max + c.rand.number(1,10))
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return



        assert False


    def test_none_of(self):
        """Confirm that we can validate value does not match a preset list"""
        # min = 12
        values = ["beep","boop"]
        value = c.rand.rand()
        v = valid.NoneOf(values)
        r = v(value,"value")
        assert r == value


        incorrect_value = c.rand.option(values)
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return

        assert False

    def test_one_of(self):
        """Confirm that we can validate that a value matches in a list"""
        # min = 12
        values = ["beep","boop"]
        value = c.rand.option(values)
        v = valid.OneOf(values)
        r = v(value,"value")
        assert r == value


        incorrect_value = c.rand.rand()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return

        assert False


    def test_past_unix_timestamp(self):
        """Confirm that we can validate a past unix timestamp."""
        ct = round(datetime.now(tz=timezone.utc).timestamp())
        value = ct - (86400 * c.rand.number(1,10))
        v = valid.PastUnixDate(value)
        r = v(value,"value")
        assert r == value

        incorrect_value = ct + (86400 * c.rand.number(1,10))
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return
        assert False


    def test_phone_number(self):
        """Confirm that we can validate a phone number"""
        value = c.rand.phone()
        v = valid.PhoneNumber(value)
        r = v(value,"value")
        assert r == value


        incorrect_value = c.rand.rand()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return

        assert False

    def test_range(self):
        """Confirm that we can validate that a value is within a range."""
        min = 1
        max = 100

        value = c.rand.number(min,max)
        v = valid.Range(min,max)
        r = v(value,"value")
        assert r == value


        incorrect_value = min - c.rand.number()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            pass

        incorrect_value = max + c.rand.number()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return

        assert False

    def test_regex(self):
        """Confirm that we can validate that a value matches a regex."""
        regex = r"beep\sboop[0-9]*"

        value = "beep boop123"
        v = valid.Regex(regex)
        r = v(value,"value")
        assert r == value


        incorrect_value = c.rand.number()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return


        assert False

    def test_credit_card(self):
        """Confirm that we can validate a credit card number."""
        value = c.rand.credit_card_number()
        v = valid.CreditCardNumber()
        r = v(value,"value")
        assert r == value


        incorrect_value = c.rand.number()
        try:
            v(incorrect_value,"value")
        except Exception as e:
            return


        assert False




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
import re
import string
from typing import Iterable, OrderedDict, Pattern, Sequence, Sized, Any, TypeVar, Union
from collections import OrderedDict
from abc import ABC, abstractmethod


import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator


_T = TypeVar("_T")

class StrongPassword(Validator):

    field_name = None
    default_message = "Weak Password"


    def __init__(
        self,
        error:str=None,
        ):
        '''
            A validator that passes if the value is a strong password

            - 8 characters
            - one lower chase alpha character
            - one upper chase alpha character
            - one numeric character
            - one special character

            ----------

            Arguments
            -------------------------
            `error` {str}
                The custom error to use if this validator fails.
                Accepts `input`,`field_name`

            Return {any}
            ----------------------
            The test value is successful, raises ValidationError if it fails.

            Meta
            ----------
            `author`: Colemen Atwood
            `created`: 03-27-2023 09:10:41
            `memberOf`: NoneOf
            `version`: 1.0
            `method_name`: NoneOf
            * @xxx [03-27-2023 09:13:12]: documentation for NoneOf
        '''
        self.regex = re.compile(r"^(?=.*?[A-Z])(?=.*?[a-z])(?=.*?[0-9])(?=.*?[#?!@$ %^&*-]).{8,}$")
        self.error = error or self.default_message  # type: str


    def _format_error(self, value: Union[str,bytes]) -> str:
        args = {
            "input":value,
        }
        if self.field_name is not None:
            args["field_name"] = self.field_name

        return self.error.format(**args)


    def _repr_args(self) -> str:
        return f""

    def __call__(self,value:str,field_name:str=None)->_T:
        self.field_name = field_name
        if self.regex.match(value) is None:
            string.ascii_lowercase
            raise ValidationError(self._format_error(value))


        return value




# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from dataclasses import dataclass
import re
from typing import Iterable, OrderedDict, Sized, Union
from collections import OrderedDict

import colemen_utils as c


class VolentError(Exception):
    '''Base Class for Volent exceptions'''



class ValidationError(VolentError):
    _errors = {}
    def __init__(
        self,
        message:Union[str,list,dict],
        field_name:str=None,
        ):
        if field_name is not None:
            self._errors[field_name] = message
        super().__init__(message)

    @property
    def errors(self):
        '''Get a dictionary of errors

        {
            field_name : error_message
        }
        '''
        if len(list(self._errors.keys())) == 0:
            data = {}
            for idx,k in enumerate(self.args):
                data[f"v{idx}"] = k
            return data
        return self._errors

class DuplicateEntryError(VolentError):
    _error = None
    _constraint_name = None
    _value = None
    _message = None
    _errno = None
    _sql_state = None
    _args = None
    def __init__(
        self,
        error,
        ):
        self._error = error
        message=self._message = error.msg
        self._errno = error.errno
        self._args = error.args
        self._sql_state = error.sqlstate

        match = re.findall(r"key\s*'([a-zA-Z0-9_]*)",message)
        if len(match) == 1:
            self._constraint_name = match[0]
        val = message.replace("Duplicate entry ","")
        val = re.sub(r"for key .*","",val)
        self._value = c.string.strip(val,["'"])
        super().__init__(self._message)

    @property
    def constraint_name(self):
        '''
            Get this exceptions's constraint_name

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-27-2023 14:37:30
            `@memberOf`: exceptions
            `@property`: constraint_name
        '''
        return self._constraint_name

    @property
    def value(self):
        '''
            Get this exceptions's value

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-27-2023 14:43:42
            `@memberOf`: exceptions
            `@property`: value
        '''
        return self._value

    @property
    def error_number(self):
        '''
            Get this exceptions's error_number

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-27-2023 15:54:29
            `@memberOf`: exceptions
            `@property`: error_number
        '''
        return self._errno

    @property
    def sql_state(self):
        '''
            Get this exceptions's sql_state

            `default`:None


            Meta
            ----------
            `@author`: Colemen Atwood
            `@created`: 04-27-2023 15:54:52
            `@memberOf`: exceptions
            `@property`: sql_state
        '''
        return self._sql_state





# # pylint: disable=missing-function-docstring
# # pylint: disable=missing-class-docstring
# # pylint: disable=line-too-long
# # pylint: disable=unused-import

# from dataclasses import dataclass
# from string import Template
# from typing import Iterable, Union
# # from typing import Iterable

# import colemen_utils as c
# import sqlite3
# import mysql.connector as _mysqlConnnector


# import volent.settings as _settings
# import volent.settings.types as _t

# qc:str = _settings.control.sql_quote_char
# '''The quote character to use.'''
# indent:str = " " * _settings.control.mysql.create_table_indent
# '''The indentation string used for the table contents'''


# @dataclass
# class ReverseEngineerMixin:
    
    
#     def gen_python_create(self,data:dict):
        
#     def gen_create(self):
#         from volent.Database import Database as _database
#         from volent.Model import Model as _model
#         from volent.Column import Column as _column
#         from volent.UniqueConstraint import UniqueConstraint as _unique_constraint
#         from volent.Relationship import Relationship as _relationship
        
#         if isinstance(self,_column):
#             value = self._gen_column_create()





# COLUMN_CREATE_TEMPLATE = """self.$name = Column($data_type()$args)"""
# def _gen_column_create(self,col:_t.column_type):
#     s = Template(COLUMN_CREATE_TEMPLATE)
#     args = []
#     if col.is_foreign_key is True:
#         args.append("is_foreign_key=True")
#     if col.is_primary is True:
#         args.append("is_primary=True")
#     if col.is_primary is True:
#         args.append("is_primary=True")
#     val = s.substitute(
#         indent=indent,
#         name=col.name,
#         data_type=col.data_type.__class__.__name__,
#     )


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

# from dataclasses import dataclass
# from typing import Iterable, OrderedDict, Union
# from collections import OrderedDict
# from abc import ABC, abstractmethod
import re

import colemen_utils as c

import volent.settings as _settings
import volent.settings.types as _t
from volent.exceptions import ValidationError
from volent.validate.Validator import Validator

class Email(Validator):

    field_name = None
    default_message = "Invalid email address."

    USER_REGEX = re.compile(
        r"(^[-!#$%&'*+/=?^`{}|~\w]+(\.[-!#$%&'*+/=?^`{}|~\w]+)*\Z"  # dot-atom
        # quoted-string
        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]'
        r'|\\[\001-\011\013\014\016-\177])*"\Z)',
        re.IGNORECASE | re.UNICODE,
    )
    
    DOMAIN_REGEX = re.compile(
        # domain
        r"(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+" r"(?:[A-Z]{2,6}|[A-Z0-9-]{2,})\Z"
        # literal form, ipv4 address (SMTP 4.1.3)
        r"|^\[(25[0-5]|2[0-4]\d|[0-1]?\d?\d)"
        r"(\.(25[0-5]|2[0-4]\d|[0-1]?\d?\d)){3}\]\Z",
        re.IGNORECASE | re.UNICODE,
    )

    DOMAIN_WHITELIST = ("localhost",)

    def __init__(
        self,
        ):
        pass
        # if equal is not None and any([min, max]):
        #     raise ValueError(
        #         "The `equal` parameter was provided, maximum or "
        #         "minimum parameter must not be provided."
        #     )

        # self.min = min
        # self.max = max
        # self.equal = equal

    def _format_error(self, message: str) -> str:
        if self.field_name is not None:
            return  f"{self.field_name} is not a valid email address"
        return self.default_message

    # def _repr_args(self) -> str:
    #     return f"min={self.min!r}, max={self.max!r}, equal={self.equal!r}"

    def __call__(self, value: str,field_name:str=None) -> str:
        self.field_name = field_name
        message = self._format_error(value)

        if not value or "@" not in value:
            raise ValidationError(message)

        user_part, domain_part = value.rsplit("@", 1)

        if not self.USER_REGEX.match(user_part):
            raise ValidationError(message)

        if domain_part not in self.DOMAIN_WHITELIST:
            if not self.DOMAIN_REGEX.match(domain_part):
                try:
                    domain_part = domain_part.encode("idna").decode("ascii")
                except UnicodeError:
                    pass
                else:
                    if self.DOMAIN_REGEX.match(domain_part):
                        return value
                raise ValidationError(message)

        return value


# pylint: disable=missing-function-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=line-too-long
# pylint: disable=unused-import

from volent.mixins.OrderedClass import OrderedClass
from volent.mixins.MySQLGeneratorMixin import MySQLGeneratorMixin
from volent.mixins.DatabaseConnection import DatabaseConnection


